type 'a trie = {
  mutable v : 'a option;
  mutable brs : (int * 'a trie) list
}

type 'a t =
  | Len of int * 'a * 'a t
  | Fork of int * int * int * 'a t array * 'a
  | Ret of 'a
  | RetIf of string * 'a * 'a
  | ForkSafe of int * int * 'a t array
  | Fail

let rec find s = function
  | Len (i,v,m) -> if String.length s = i then v else find s m
  | Fork (i,ofs,len,ar,v) ->
      let c = Char.code s.[i] - ofs in
      if (c < 0) || (c >= len) then v
      else find s ar.(c)
  | ForkSafe (i,ofs,ar) ->
      let c = Char.code s.[i] - ofs in
      find s ar.(c)
  | Ret v -> v
  | RetIf (s',v,v') -> if s = s' then v else v'
  | Fail -> assert false

let rec map_of_trie t i =
  let m =
    match t.brs with
      | [] -> Fail
      | [ (c,t') ] -> map_of_trie t' (succ i)
      | brs ->
	  let codes = List.map fst brs in
	  let min = List.fold_left min 256 codes
	  and max = List.fold_left max (-1) codes in
	  let ar = Array.create (max - min + 1) Fail in
	  List.iter (fun (c,t') -> ar.(c - min) <- map_of_trie t' (succ i)) brs;
	  ForkSafe (i,min,ar)
  in
  match t.v with
    | Some v -> (match m with Fail -> Ret v | m -> Len (i,v,m))
    | None -> m

let rec map_of_trie_def def t i =
  let m =
    match t.brs with
      | [] -> Ret def
      | brs ->
	  let codes = List.map fst brs in
	  let min = List.fold_left min 256 codes
	  and max = List.fold_left max (-1) codes in
	  let ar = Array.create (max - min + 1) (Ret def) in
	  List.iter (fun (c,t') -> ar.(c - min) <- map_of_trie_def def t' (succ i)) brs;
	  Fork (i,min,max-min+1,ar,def)
  in
  match t.v with
    | Some v -> Len (i,v,m)
    | None -> Len (i,def,m)


let empty () = { v = None; brs = [] }

let rec add t k i v =
  if (i = String.length k) then t.v <- Some v
  else
    let c = Char.code k.[i] in
    let t' =  
      try List.assq c t.brs
      with Not_found -> 
	let t' = empty () in
	t.brs <- (c,t') :: t.brs;
	t' in
    add t' k (succ i) v
  

let prepare def lst =
  let trie = empty () in
  List.iter (fun (k,v) -> add trie k 0 v) lst;
  match def with None -> map_of_trie trie 0 | Some d -> map_of_trie_def d trie 0
